#include "common.h"
#include <cstdio>
#include <cstdlib>

using namespace v8;
using namespace node;

#include <iostream>
using namespace std;

namespace glfw {

/* @Module: initialization and version information */

JS_METHOD(Init) {
  HandleScope scope;
  return scope.Close(JS_BOOL(glfwInit()==GL_TRUE));
}

JS_METHOD(Terminate) {
  HandleScope scope;
  glfwTerminate();
  return scope.Close(Undefined());
}

JS_METHOD(GetVersion) {
  HandleScope scope;
  int major, minor, rev;
  glfwGetVersion(&major,&minor,&rev);
  Local<Array> arr=Array::New(3);
  arr->Set(JS_STR("major"),JS_INT(major));
  arr->Set(JS_STR("minor"),JS_INT(minor));
  arr->Set(JS_STR("rev"),JS_INT(rev));
  return scope.Close(arr);
}

JS_METHOD(GetVersionString) {
  HandleScope scope;
  const char* ver=glfwGetVersionString();
  return scope.Close(JS_STR(ver));
}

/* @Module: Time input */

JS_METHOD(GetTime) {
  HandleScope scope;
  return scope.Close(JS_NUM(glfwGetTime()));
}

JS_METHOD(SetTime) {
  HandleScope scope;
  double time = args[0]->NumberValue();
  glfwSetTime(time);
  return scope.Close(Undefined());
}

/* @Module: monitor handling */

/* TODO: Monitor configuration change callback */

JS_METHOD(GetMonitors) {
  HandleScope scope;
  int monitor_count, mode_count, xpos, ypos, width, height;
  int i, j;
  GLFWmonitor **monitors = glfwGetMonitors(&monitor_count);
  GLFWmonitor *primary = glfwGetPrimaryMonitor();
  const GLFWvidmode *mode, *modes;
  
  Local<Array> js_monitors = Array::New(monitor_count);
  Local<Object> js_monitor, js_mode;
  Local<Array> js_modes;
  for(i=0; i<monitor_count; i++){
    js_monitor = Object::New();
    js_monitor->Set(JS_STR("is_primary"), JS_BOOL(monitors[i] == primary));
    js_monitor->Set(JS_STR("index"), JS_INT(i));
    
    js_monitor->Set(JS_STR("name"), JS_STR(glfwGetMonitorName(monitors[i])));
    
    glfwGetMonitorPos(monitors[i], &xpos, &ypos);
    js_monitor->Set(JS_STR("pos_x"), JS_INT(xpos));
    js_monitor->Set(JS_STR("pos_y"), JS_INT(ypos));
    
    glfwGetMonitorPhysicalSize(monitors[i], &width, &height);
    js_monitor->Set(JS_STR("width_mm"), JS_INT(width));
    js_monitor->Set(JS_STR("height_mm"), JS_INT(height));
    
    mode = glfwGetVideoMode(monitors[i]);
    js_monitor->Set(JS_STR("width"), JS_INT(mode->width));
    js_monitor->Set(JS_STR("height"), JS_INT(mode->height));
    js_monitor->Set(JS_STR("rate"), JS_INT(mode->refreshRate));
    
    modes = glfwGetVideoModes(monitors[i], &mode_count);
    js_modes = Array::New(mode_count);
    for(j=0; j<mode_count; j++){
      js_mode = Object::New();
      js_mode->Set(JS_STR("width"), JS_INT(modes[j].width));
      js_mode->Set(JS_STR("height"), JS_INT(modes[j].height));
      js_mode->Set(JS_STR("rate"), JS_INT(modes[j].refreshRate));
      // NOTE: Are color bits necessary?
      js_modes->Set(JS_INT(j), js_mode);
    }
    js_monitor->Set(JS_STR("modes"), js_modes);
    
    js_monitors->Set(JS_INT(i), js_monitor);
  }
  
  return scope.Close(js_monitors);
}


/* @Module: Window handling */
Persistent<Object> glfw_events;
int lastX=0,lastY=0;
bool windowCreated=false;

JS_METHOD(testScene) {
  HandleScope scope;
  int width = args[0]->Uint32Value();
  int height = args[1]->Uint32Value();
  float ratio = width / (float) height;

  glViewport(0, 0, width, height);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-ratio, ratio, -1.f, 1.f, 1.f, -1.f);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glRotatef((float) glfwGetTime() * 50.f, 0.f, 0.f, 1.f);

  glBegin(GL_TRIANGLES);
  glColor3f(1.f, 0.f, 0.f);
  glVertex3f(-0.6f, -0.4f, 0.f);
  glColor3f(0.f, 1.f, 0.f);
  glVertex3f(0.6f, -0.4f, 0.f);
  glColor3f(0.f, 0.f, 1.f);
  glVertex3f(0.f, 0.6f, 0.f);
  glEnd();

  return scope.Close(Undefined());
}

JS_METHOD(WindowHint) {
  HandleScope scope;
  int target       = args[0]->Uint32Value();
  int hint         = args[1]->Uint32Value();
  glfwWindowHint(target, hint);
  return scope.Close(Undefined());
}

JS_METHOD(DefaultWindowHints) {
  HandleScope scope;
  glfwDefaultWindowHints();
  return scope.Close(Undefined());
}

JS_METHOD(CreateWindow) {
  HandleScope scope;
  int width       = args[0]->Uint32Value();
  int height      = args[1]->Uint32Value();
  String::Utf8Value str(args[2]->ToString());
  int monitor_idx = args[3]->Uint32Value();
  
  GLFWwindow* window = NULL;
  GLFWmonitor **monitors = NULL, *monitor = NULL;
  int monitor_count;
  if(args.Length() >= 4 && monitor_idx >= 0){
    monitors = glfwGetMonitors(&monitor_count);
    if(monitor_idx >= monitor_count){
      return ThrowError("Invalid monitor");
    }
    monitor = monitors[monitor_idx];
  }

  if(!windowCreated) {
    window = glfwCreateWindow(width, height, *str, monitor, NULL);
    if(!window) {
      // can't create window, throw error
      return ThrowError("Can't create GLFW window");
    }

    glfwMakeContextCurrent(window);

    GLenum err = glewInit();
    if (err)
    {
      /* Problem: glewInit failed, something is seriously wrong. */
      string msg="Can't init GLEW (glew error ";
      msg+=(const char*) glewGetErrorString(err);
      msg+=")";

      fprintf(stderr, "%s", msg.c_str());
      return ThrowError(msg.c_str());
    }
    fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));
  }
  else
    glfwSetWindowSize(window, width,height);

  // Set callback functions
  glfw_events=Persistent<Object>::New(args.This()->Get(JS_STR("events"))->ToObject());

  return scope.Close(JS_NUM((uint64_t) window));
}

JS_METHOD(DestroyWindow) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwDestroyWindow(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(SetWindowTitle) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  String::Utf8Value str(args[1]->ToString());
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSetWindowTitle(window, *str);
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetWindowSize) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    int w,h;
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwGetWindowSize(window, &w, &h);
    Local<Array> arr=Array::New(2);
    arr->Set(JS_STR("width"),JS_INT(w));
    arr->Set(JS_STR("height"),JS_INT(h));
    return scope.Close(arr);
  }
  return scope.Close(Undefined());
}

JS_METHOD(SetWindowSize) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSetWindowSize(window, args[1]->Uint32Value(),args[2]->Uint32Value());
  }
  return scope.Close(Undefined());
}

JS_METHOD(SetWindowPos) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSetWindowPos(window, args[1]->Uint32Value(),args[2]->Uint32Value());
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetWindowPos) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    int xpos, ypos;
    glfwGetWindowPos(window, &xpos, &ypos);
    Local<Array> arr=Array::New(2);
    arr->Set(JS_STR("xpos"),JS_INT(xpos));
    arr->Set(JS_STR("ypos"),JS_INT(ypos));
    return scope.Close(arr);
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetFramebufferSize) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    int width, height;
    glfwGetFramebufferSize(window, &width, &height);
    Local<Array> arr=Array::New(2);
    arr->Set(JS_STR("width"),JS_INT(width));
    arr->Set(JS_STR("height"),JS_INT(height));
    return scope.Close(arr);
  }
  return scope.Close(Undefined());
}

JS_METHOD(IconifyWindow) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwIconifyWindow(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(RestoreWindow) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwRestoreWindow(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(HideWindow) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwHideWindow(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(ShowWindow) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwShowWindow(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(WindowShouldClose) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    return scope.Close(JS_INT(glfwWindowShouldClose(window)));
  }
  return scope.Close(Undefined());
}

JS_METHOD(SetWindowShouldClose) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  int value=args[1]->Uint32Value();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSetWindowShouldClose(window, value);
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetWindowAttrib) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  int attrib=args[1]->Uint32Value();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    return scope.Close(JS_INT(glfwGetWindowAttrib(window, attrib)));
  }
  return scope.Close(Undefined());
}

JS_METHOD(PollEvents) {
  HandleScope scope;
  glfwPollEvents();
  return scope.Close(Undefined());
}

JS_METHOD(WaitEvents) {
  HandleScope scope;
  glfwWaitEvents();
  return scope.Close(Undefined());
}

//GLFWAPI void GLFWAPIENTRY glfwSetWindowSizeCallback( GLFWwindowsizefun cbfun );
//GLFWAPI void GLFWAPIENTRY glfwSetWindowCloseCallback( GLFWwindowclosefun cbfun );
//GLFWAPI void GLFWAPIENTRY glfwSetWindowRefreshCallback( GLFWwindowrefreshfun cbfun );

/* Input handling */

JS_METHOD(GetKey) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  int key=args[1]->Uint32Value();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    return scope.Close(JS_INT(glfwGetKey(window, key)));
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetMouseButton) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  int button=args[1]->Uint32Value();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    return scope.Close(JS_INT(glfwGetMouseButton(window, button)));
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetCursorPos) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    double x,y;
    glfwGetCursorPos(window, &x, &y);
    Local<Array> arr=Array::New(2);
    arr->Set(JS_STR("x"),JS_INT(x));
    arr->Set(JS_STR("y"),JS_INT(y));
    return scope.Close(arr);
  }
  return scope.Close(Undefined());
}

JS_METHOD(SetCursorPos) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  int x=args[1]->NumberValue();
  int y=args[2]->NumberValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSetCursorPos(window, x, y);
  }
  return scope.Close(Undefined());
}

/* @Module Context handling */
JS_METHOD(MakeContextCurrent) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwMakeContextCurrent(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(GetCurrentContext) {
  HandleScope scope;
  GLFWwindow* window=glfwGetCurrentContext();
  return scope.Close(JS_NUM((uint64_t) window));
}

JS_METHOD(SwapBuffers) {
  HandleScope scope;
  uint64_t handle=args[0]->IntegerValue();
  if(handle) {
    GLFWwindow* window = reinterpret_cast<GLFWwindow*>(handle);
    glfwSwapBuffers(window);
  }
  return scope.Close(Undefined());
}

JS_METHOD(SwapInterval) {
  HandleScope scope;
  int interval=args[0]->Int32Value();
  glfwSwapInterval(interval);
  return scope.Close(Undefined());
}

JS_METHOD(ExtensionSupported) {
  HandleScope scope;
  String::AsciiValue str(args[0]->ToString());
  return scope.Close(JS_BOOL(glfwExtensionSupported(*str)==1));
}

// make sure we close everything when we exit
void AtExit() {
  glfwTerminate();
}

} // namespace glfw

///////////////////////////////////////////////////////////////////////////////
//
// bindings
//
///////////////////////////////////////////////////////////////////////////////
#define JS_GLFW_CONSTANT(name) target->Set(JS_STR( #name ), JS_INT(GLFW_ ## name))
#define JS_GLFW_SET_METHOD(name) NODE_SET_METHOD(target, #name , glfw::name);

extern "C" {
void init(Handle<Object> target) {
  atexit(glfw::AtExit);

  HandleScope scope;

  /* GLFW initialization, termination and version querying */
  JS_GLFW_SET_METHOD(Init);
  JS_GLFW_SET_METHOD(Terminate);
  JS_GLFW_SET_METHOD(GetVersion);
  JS_GLFW_SET_METHOD(GetVersionString);

  /* Time */
  JS_GLFW_SET_METHOD(GetTime);
  JS_GLFW_SET_METHOD(SetTime);
  
  /* Monitor handling */
  JS_GLFW_SET_METHOD(GetMonitors);

  /* Window handling */
  JS_GLFW_SET_METHOD(CreateWindow);
  JS_GLFW_SET_METHOD(WindowHint);
  JS_GLFW_SET_METHOD(DefaultWindowHints);
  JS_GLFW_SET_METHOD(DestroyWindow);
  JS_GLFW_SET_METHOD(SetWindowShouldClose);
  JS_GLFW_SET_METHOD(WindowShouldClose);
  JS_GLFW_SET_METHOD(SetWindowTitle);
  JS_GLFW_SET_METHOD(GetWindowSize);
  JS_GLFW_SET_METHOD(SetWindowSize);
  JS_GLFW_SET_METHOD(SetWindowPos);
  JS_GLFW_SET_METHOD(GetWindowPos);
  JS_GLFW_SET_METHOD(GetFramebufferSize);
  JS_GLFW_SET_METHOD(IconifyWindow);
  JS_GLFW_SET_METHOD(RestoreWindow);
  JS_GLFW_SET_METHOD(ShowWindow);
  JS_GLFW_SET_METHOD(HideWindow);
  JS_GLFW_SET_METHOD(GetWindowAttrib);
  JS_GLFW_SET_METHOD(PollEvents);
  JS_GLFW_SET_METHOD(WaitEvents);

  /* Input handling */
  JS_GLFW_SET_METHOD(GetKey);
  JS_GLFW_SET_METHOD(GetMouseButton);
  JS_GLFW_SET_METHOD(GetCursorPos);
  JS_GLFW_SET_METHOD(SetCursorPos);

  /* Context handling */
  JS_GLFW_SET_METHOD(MakeContextCurrent);
  JS_GLFW_SET_METHOD(GetCurrentContext);
  JS_GLFW_SET_METHOD(SwapBuffers);
  JS_GLFW_SET_METHOD(SwapInterval);
  JS_GLFW_SET_METHOD(ExtensionSupported);

  /*************************************************************************
   * GLFW version
   *************************************************************************/

  JS_GLFW_CONSTANT(VERSION_MAJOR);
  JS_GLFW_CONSTANT(VERSION_MINOR);
  JS_GLFW_CONSTANT(VERSION_REVISION);


  /*************************************************************************
   * Input handling definitions
   *************************************************************************/

  /* Key and button state/action definitions */
  JS_GLFW_CONSTANT(RELEASE);
  JS_GLFW_CONSTANT(PRESS);
  JS_GLFW_CONSTANT(REPEAT);

  /* These key codes are inspired by the *USB HID Usage Tables v1.12* (p. 53-60),
   * but re-arranged to map to 7-bit ASCII for printable keys (function keys are
   * put in the 256+ range).
   *
   * The naming of the key codes follow these rules:
   *  - The US keyboard layout is used
   *  - Names of printable alpha-numeric characters are used (e.g. "A", "R",
   *    "3", etc.)
   *  - For non-alphanumeric characters, Unicode:ish names are used (e.g.
   *    "COMMA", "LEFT_SQUARE_BRACKET", etc.). Note that some names do not
   *    correspond to the Unicode standard (usually for brevity)
   *  - Keys that lack a clear US mapping are named "WORLD_x"
   *  - For non-printable keys, custom names are used (e.g. "F4",
   *    "BACKSPACE", etc.)
   */

  /* The unknown key */
  JS_GLFW_CONSTANT(KEY_UNKNOWN);

  /* Printable keys */
  JS_GLFW_CONSTANT(KEY_SPACE);
  JS_GLFW_CONSTANT(KEY_APOSTROPHE);
  JS_GLFW_CONSTANT(KEY_COMMA);
  JS_GLFW_CONSTANT(KEY_MINUS);
  JS_GLFW_CONSTANT(KEY_PERIOD);
  JS_GLFW_CONSTANT(KEY_SLASH);
  JS_GLFW_CONSTANT(KEY_0);
  JS_GLFW_CONSTANT(KEY_1);
  JS_GLFW_CONSTANT(KEY_2);
  JS_GLFW_CONSTANT(KEY_3);
  JS_GLFW_CONSTANT(KEY_4);
  JS_GLFW_CONSTANT(KEY_5);
  JS_GLFW_CONSTANT(KEY_6);
  JS_GLFW_CONSTANT(KEY_7);
  JS_GLFW_CONSTANT(KEY_8);
  JS_GLFW_CONSTANT(KEY_9);
  JS_GLFW_CONSTANT(KEY_SEMICOLON);
  JS_GLFW_CONSTANT(KEY_EQUAL);
  JS_GLFW_CONSTANT(KEY_A);
  JS_GLFW_CONSTANT(KEY_B);
  JS_GLFW_CONSTANT(KEY_C);
  JS_GLFW_CONSTANT(KEY_D);
  JS_GLFW_CONSTANT(KEY_E);
  JS_GLFW_CONSTANT(KEY_F);
  JS_GLFW_CONSTANT(KEY_G);
  JS_GLFW_CONSTANT(KEY_H);
  JS_GLFW_CONSTANT(KEY_I);
  JS_GLFW_CONSTANT(KEY_J);
  JS_GLFW_CONSTANT(KEY_K);
  JS_GLFW_CONSTANT(KEY_L);
  JS_GLFW_CONSTANT(KEY_M);
  JS_GLFW_CONSTANT(KEY_N);
  JS_GLFW_CONSTANT(KEY_O);
  JS_GLFW_CONSTANT(KEY_P);
  JS_GLFW_CONSTANT(KEY_Q);
  JS_GLFW_CONSTANT(KEY_R);
  JS_GLFW_CONSTANT(KEY_S);
  JS_GLFW_CONSTANT(KEY_T);
  JS_GLFW_CONSTANT(KEY_U);
  JS_GLFW_CONSTANT(KEY_V);
  JS_GLFW_CONSTANT(KEY_W);
  JS_GLFW_CONSTANT(KEY_X);
  JS_GLFW_CONSTANT(KEY_Y);
  JS_GLFW_CONSTANT(KEY_Z);
  JS_GLFW_CONSTANT(KEY_LEFT_BRACKET);
  JS_GLFW_CONSTANT(KEY_BACKSLASH);
  JS_GLFW_CONSTANT(KEY_RIGHT_BRACKET);
  JS_GLFW_CONSTANT(KEY_GRAVE_ACCENT);
  JS_GLFW_CONSTANT(KEY_WORLD_1);
  JS_GLFW_CONSTANT(KEY_WORLD_2);

  /* Function keys */
  JS_GLFW_CONSTANT(KEY_ESCAPE);
  JS_GLFW_CONSTANT(KEY_ENTER);
  JS_GLFW_CONSTANT(KEY_TAB);
  JS_GLFW_CONSTANT(KEY_BACKSPACE);
  JS_GLFW_CONSTANT(KEY_INSERT);
  JS_GLFW_CONSTANT(KEY_DELETE);
  JS_GLFW_CONSTANT(KEY_RIGHT);
  JS_GLFW_CONSTANT(KEY_LEFT);
  JS_GLFW_CONSTANT(KEY_DOWN);
  JS_GLFW_CONSTANT(KEY_UP);
  JS_GLFW_CONSTANT(KEY_PAGE_UP);
  JS_GLFW_CONSTANT(KEY_PAGE_DOWN);
  JS_GLFW_CONSTANT(KEY_HOME);
  JS_GLFW_CONSTANT(KEY_END);
  JS_GLFW_CONSTANT(KEY_CAPS_LOCK);
  JS_GLFW_CONSTANT(KEY_SCROLL_LOCK);
  JS_GLFW_CONSTANT(KEY_NUM_LOCK);
  JS_GLFW_CONSTANT(KEY_PRINT_SCREEN);
  JS_GLFW_CONSTANT(KEY_PAUSE);
  JS_GLFW_CONSTANT(KEY_F1);
  JS_GLFW_CONSTANT(KEY_F2);
  JS_GLFW_CONSTANT(KEY_F3);
  JS_GLFW_CONSTANT(KEY_F4);
  JS_GLFW_CONSTANT(KEY_F5);
  JS_GLFW_CONSTANT(KEY_F6);
  JS_GLFW_CONSTANT(KEY_F7);
  JS_GLFW_CONSTANT(KEY_F8);
  JS_GLFW_CONSTANT(KEY_F9);
  JS_GLFW_CONSTANT(KEY_F10);
  JS_GLFW_CONSTANT(KEY_F11);
  JS_GLFW_CONSTANT(KEY_F12);
  JS_GLFW_CONSTANT(KEY_F13);
  JS_GLFW_CONSTANT(KEY_F14);
  JS_GLFW_CONSTANT(KEY_F15);
  JS_GLFW_CONSTANT(KEY_F16);
  JS_GLFW_CONSTANT(KEY_F17);
  JS_GLFW_CONSTANT(KEY_F18);
  JS_GLFW_CONSTANT(KEY_F19);
  JS_GLFW_CONSTANT(KEY_F20);
  JS_GLFW_CONSTANT(KEY_F21);
  JS_GLFW_CONSTANT(KEY_F22);
  JS_GLFW_CONSTANT(KEY_F23);
  JS_GLFW_CONSTANT(KEY_F24);
  JS_GLFW_CONSTANT(KEY_F25);
  JS_GLFW_CONSTANT(KEY_KP_0);
  JS_GLFW_CONSTANT(KEY_KP_1);
  JS_GLFW_CONSTANT(KEY_KP_2);
  JS_GLFW_CONSTANT(KEY_KP_3);
  JS_GLFW_CONSTANT(KEY_KP_4);
  JS_GLFW_CONSTANT(KEY_KP_5);
  JS_GLFW_CONSTANT(KEY_KP_6);
  JS_GLFW_CONSTANT(KEY_KP_7);
  JS_GLFW_CONSTANT(KEY_KP_8);
  JS_GLFW_CONSTANT(KEY_KP_9);
  JS_GLFW_CONSTANT(KEY_KP_DECIMAL);
  JS_GLFW_CONSTANT(KEY_KP_DIVIDE);
  JS_GLFW_CONSTANT(KEY_KP_MULTIPLY);
  JS_GLFW_CONSTANT(KEY_KP_SUBTRACT);
  JS_GLFW_CONSTANT(KEY_KP_ADD);
  JS_GLFW_CONSTANT(KEY_KP_ENTER);
  JS_GLFW_CONSTANT(KEY_KP_EQUAL);
  JS_GLFW_CONSTANT(KEY_LEFT_SHIFT);
  JS_GLFW_CONSTANT(KEY_LEFT_CONTROL);
  JS_GLFW_CONSTANT(KEY_LEFT_ALT);
  JS_GLFW_CONSTANT(KEY_LEFT_SUPER);
  JS_GLFW_CONSTANT(KEY_RIGHT_SHIFT);
  JS_GLFW_CONSTANT(KEY_RIGHT_CONTROL);
  JS_GLFW_CONSTANT(KEY_RIGHT_ALT);
  JS_GLFW_CONSTANT(KEY_RIGHT_SUPER);
  JS_GLFW_CONSTANT(KEY_MENU);
  JS_GLFW_CONSTANT(KEY_LAST);

  /*Modifier key flags*/

  /*If this bit is set one or more Shift keys were held down. */
  JS_GLFW_CONSTANT(MOD_SHIFT);
  /*If this bit is set one or more Control keys were held down. */
  JS_GLFW_CONSTANT(MOD_CONTROL);
  /*If this bit is set one or more Alt keys were held down. */
  JS_GLFW_CONSTANT(MOD_ALT);
  /*If this bit is set one or more Super keys were held down. */
  JS_GLFW_CONSTANT(MOD_SUPER);

  /*Mouse buttons*/
  JS_GLFW_CONSTANT(MOUSE_BUTTON_1);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_2);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_3);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_4);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_5);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_6);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_7);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_8);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_LAST);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_LEFT);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_RIGHT);
  JS_GLFW_CONSTANT(MOUSE_BUTTON_MIDDLE);

  /*Joysticks*/
  JS_GLFW_CONSTANT(JOYSTICK_1);
  JS_GLFW_CONSTANT(JOYSTICK_2);
  JS_GLFW_CONSTANT(JOYSTICK_3);
  JS_GLFW_CONSTANT(JOYSTICK_4);
  JS_GLFW_CONSTANT(JOYSTICK_5);
  JS_GLFW_CONSTANT(JOYSTICK_6);
  JS_GLFW_CONSTANT(JOYSTICK_7);
  JS_GLFW_CONSTANT(JOYSTICK_8);
  JS_GLFW_CONSTANT(JOYSTICK_9);
  JS_GLFW_CONSTANT(JOYSTICK_10);
  JS_GLFW_CONSTANT(JOYSTICK_11);
  JS_GLFW_CONSTANT(JOYSTICK_12);
  JS_GLFW_CONSTANT(JOYSTICK_13);
  JS_GLFW_CONSTANT(JOYSTICK_14);
  JS_GLFW_CONSTANT(JOYSTICK_15);
  JS_GLFW_CONSTANT(JOYSTICK_16);
  JS_GLFW_CONSTANT(JOYSTICK_LAST);

  /*errors Error codes*/

  /*GLFW has not been initialized.*/
  JS_GLFW_CONSTANT(NOT_INITIALIZED);
  /*No context is current for this thread.*/
  JS_GLFW_CONSTANT(NO_CURRENT_CONTEXT);
  /*One of the enum parameters for the function was given an invalid enum.*/
  JS_GLFW_CONSTANT(INVALID_ENUM);
  /*One of the parameters for the function was given an invalid value.*/
  JS_GLFW_CONSTANT(INVALID_VALUE);
  /*A memory allocation failed.*/
  JS_GLFW_CONSTANT(OUT_OF_MEMORY);
  /*GLFW could not find support for the requested client API on the system.*/
  JS_GLFW_CONSTANT(API_UNAVAILABLE);
  /*The requested client API version is not available.*/
  JS_GLFW_CONSTANT(VERSION_UNAVAILABLE);
  /*A platform-specific error occurred that does not match any of the more specific categories.*/
  JS_GLFW_CONSTANT(PLATFORM_ERROR);
  /*The clipboard did not contain data in the requested format.*/
  JS_GLFW_CONSTANT(FORMAT_UNAVAILABLE);

  JS_GLFW_CONSTANT(FOCUSED);
  JS_GLFW_CONSTANT(ICONIFIED);
  JS_GLFW_CONSTANT(RESIZABLE);
  JS_GLFW_CONSTANT(VISIBLE);
  JS_GLFW_CONSTANT(DECORATED);

  JS_GLFW_CONSTANT(RED_BITS);
  JS_GLFW_CONSTANT(GREEN_BITS);
  JS_GLFW_CONSTANT(BLUE_BITS);
  JS_GLFW_CONSTANT(ALPHA_BITS);
  JS_GLFW_CONSTANT(DEPTH_BITS);
  JS_GLFW_CONSTANT(STENCIL_BITS);
  JS_GLFW_CONSTANT(ACCUM_RED_BITS);
  JS_GLFW_CONSTANT(ACCUM_GREEN_BITS);
  JS_GLFW_CONSTANT(ACCUM_BLUE_BITS);
  JS_GLFW_CONSTANT(ACCUM_ALPHA_BITS);
  JS_GLFW_CONSTANT(AUX_BUFFERS);
  JS_GLFW_CONSTANT(STEREO);
  JS_GLFW_CONSTANT(SAMPLES);
  JS_GLFW_CONSTANT(SRGB_CAPABLE);
  JS_GLFW_CONSTANT(REFRESH_RATE);

  JS_GLFW_CONSTANT(CLIENT_API);
  JS_GLFW_CONSTANT(CONTEXT_VERSION_MAJOR);
  JS_GLFW_CONSTANT(CONTEXT_VERSION_MINOR);
  JS_GLFW_CONSTANT(CONTEXT_REVISION);
  JS_GLFW_CONSTANT(CONTEXT_ROBUSTNESS);
  JS_GLFW_CONSTANT(OPENGL_FORWARD_COMPAT);
  JS_GLFW_CONSTANT(OPENGL_DEBUG_CONTEXT);
  JS_GLFW_CONSTANT(OPENGL_PROFILE);

  JS_GLFW_CONSTANT(OPENGL_API);
  JS_GLFW_CONSTANT(OPENGL_ES_API);

  JS_GLFW_CONSTANT(NO_ROBUSTNESS);
  JS_GLFW_CONSTANT(NO_RESET_NOTIFICATION);
  JS_GLFW_CONSTANT(LOSE_CONTEXT_ON_RESET);

  JS_GLFW_CONSTANT(OPENGL_ANY_PROFILE);
  JS_GLFW_CONSTANT(OPENGL_CORE_PROFILE);
  JS_GLFW_CONSTANT(OPENGL_COMPAT_PROFILE);

  JS_GLFW_CONSTANT(CURSOR);
  JS_GLFW_CONSTANT(STICKY_KEYS);
  JS_GLFW_CONSTANT(STICKY_MOUSE_BUTTONS);

  JS_GLFW_CONSTANT(CURSOR_NORMAL);
  JS_GLFW_CONSTANT(CURSOR_HIDDEN);
  JS_GLFW_CONSTANT(CURSOR_DISABLED);

  JS_GLFW_CONSTANT(CONNECTED);
  JS_GLFW_CONSTANT(DISCONNECTED);

  JS_GLFW_SET_METHOD(testScene);
}

NODE_MODULE(glfw, init)
}

